/*
 *	Clustering algorithm for Proteinortho
 *	Reads edge list and splits connected components
 *	according to algebraic connectivity threshold
 *
 *	Last updated: 2018/03/01		
 *	Author: Marcus Lechner
 */

#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <map>
#include <algorithm>
#include <cmath>
#include <vector>
#include <stack>
#include <iomanip>
#include <cstdlib>
#include <ctime>

using namespace std;

// Structs
struct protein {vector<unsigned int> edges; unsigned int species_id; string full_name;};

// Functions
double string2double(string);
void tokenize(const string& , vector<string>& , const string&);
void parse_file(string);
void remove_edge_index(const unsigned int, const unsigned int);
double getConnectivity(vector<unsigned int>&);
void clear_edges(vector<unsigned int>&);
void partition_graph(void);
void print_group(vector<unsigned int>& , double );
float calc_group(vector<unsigned int>&);
void print_header(void);
void sort_species(void);
void stats(double, double);
void splitGroups(vector<double>&, vector<unsigned int>&);
string getTime(void);
void debug__graph_integrity(vector<unsigned int>&);
void debug__print_edgelist (protein&, const unsigned int, const int);
void debug__conn_integrity(vector<unsigned int>&, double);

// Parameters
bool param_verbose 		= false;
double param_con_threshold 	= 0.1;		// as a reference: a chain a-b-c-d has 0.25
unsigned int debug_level	= 0;
double param_sep_purity 	= 0.75;		// as a reference: a-b-c will give +/-0.707107
unsigned int param_max_nodes	= 16384;
float param_min_species	= 0;
string param_rmgraph            = "remove.graph";

// min/max number of alg con iterations
#define min_iter 16			// below this value, results may vary
#define max_iter 4096			// below this value, results may vary
#define epsilon 1e-07			// above this value, results may vary

// Globals
unsigned int species_counter = 0;	// Species
unsigned int protein_counter = 0;	// Proteins
vector<string> species;			// Number -> Name
vector<protein> graph;			// Graph containing all protein data
double last_stat = 0;			// For progress stats
unsigned int edges = 0;			// number of edges
ofstream graph_clean;			// File to store graph data
vector<int> reorder_table;		// Tells how proteins/species must be sorted

// TMP Globals
map<string,int> species2id;		// Name -> Number
map<string,int> protein2id;		// Name -> Number


///////////////////////////////////////////////////////////
// Main
///////////////////////////////////////////////////////////
void printHelp() {
	cerr << "Proteinortho5 - Spectral partitioning algorithm v5.17" << endl;
	cerr << "-----------------------------------------------------" << endl;
	cerr << "This tool is part of Proteinortho" << endl;
	cerr << "" << endl;
	cerr << "Usage:   proteinortho5_clustering [OPTIONS] graph_files..." << endl;
	cerr << "Options: -verbose          report progress" << endl;
	cerr << "         -conn double      threshold for connectivity [0.1]" << endl;
	cerr << "         -purity double    threshold for purity [0.1]" << endl;
	cerr << "         -maxnodes int     max. number of nodes for alg. clustering" << endl;
	cerr << "         -minspecies float threshold for species number [0]" << endl;
	cerr << "         -rmgraph STRING   output file for graph" << endl;
}

int main(int argc, char *argv[]) {
	if (argc <= 1) {
		printHelp();
		return EXIT_FAILURE;
	}

	try {
		// Read parameters
		int paras;
		vector<string> files;
		for (paras = 1; paras < argc; paras++) {
			string parameter = string(argv[paras]);
			if (parameter.substr(0, 1) != "-") {
				files.push_back(parameter);
			}
			else if (parameter == "-verbose") {
				paras++;
				if (string2double(string(argv[paras]))) {
					param_verbose = true;
				}
			}
			else if (parameter == "-conn") {
				paras++;
				param_con_threshold = string2double(string(argv[paras]));
			}
			else if (parameter == "-purity") {
				paras++;
				param_sep_purity = string2double(string(argv[paras]));
			}
			else if (parameter == "-maxnodes") {
				paras++;
				param_max_nodes = string2double(string(argv[paras]));
			}
			else if (parameter == "-minspecies") {
				paras++;
				param_min_species = string2double(string(argv[paras]));
				if (param_min_species < 1) {throw string("-minspecies must at least be 1. Less than one gene per species is not possible as we only count those that have an entry.");}
			}
			else if (parameter == "-debug") {
				paras++;
				debug_level = int(string2double(string(argv[paras])));
			}
			else if (parameter == "-rmgraph") {
				paras++;
				param_rmgraph = string(argv[paras]);
			}
			else {
				printHelp();
				cerr << endl << "Sorry, unknown option '" << string(argv[paras]) << "'!" << endl;
				return EXIT_FAILURE;
			}
		}

		// init RNG
//		srand(12345);		

		if (debug_level > 0) cerr << getTime() << " [DEBUG]   Debug level " << debug_level << endl;

		// Parse files
		for (vector<string>::iterator it=files.begin() ; it != files.end(); it++) {
			if (debug_level > 0) cerr << getTime() << " [DEBUG]   Parsing file " << *it << endl;
			parse_file(*it);
			if (debug_level > 0) cerr << getTime() << " [DEBUG]   I know " << species_counter <<  " species with " << protein_counter << " proteins and " << edges << " edges in sum" << endl;
		};
		// No more preallocated memory required
		graph.shrink_to_fit();

		// Free memory
		files.clear();
		species2id.clear();
		protein2id.clear();

		// Stats
		if (param_verbose) cerr << species_counter << " species" << endl << protein_counter << " paired proteins" << endl << edges << " bidirectional edges" << endl;

		if (debug_level > 0) cerr << getTime() << " [DEBUG]   Maximumum number of nodes for connectivity calculations is " << param_max_nodes << endl;

		// Prepare sort of output
		if (debug_level > 0) cerr << getTime() << " [DEBUG]   Sorting known species" << endl;
		sort_species();

		// Write output header
		print_header();							

		// Open graph-removal file
		graph_clean.open(param_rmgraph.c_str());

		// Clustering
		if (debug_level > 0) cerr << getTime() << " [DEBUG]   Clustering" << endl;
		partition_graph();
		graph_clean.close();
	}
	catch(string& error) {
		cerr << "[ERROR]   " << error << endl;
		return EXIT_FAILURE;
	}
	return EXIT_SUCCESS;
}

///////////////////////////////////////////////////////////
// Major partioning algorithm
///////////////////////////////////////////////////////////
void partition_graph() {
	vector<bool> done = vector<bool> (protein_counter, false);	// Keep track on what was done

	// For each protein (no increment here, we might redo protein i)
	for (unsigned int protein_id = 0; protein_id < graph.size();) {
//		cerr << "Protein ID: " << protein_id << endl;
		// We were here already
		if (done[protein_id])		{protein_id++;	continue;}
		if (debug_level > 0) cerr << getTime() << " [DEBUG]   Searching for connected component (ID: " << protein_id << ")" << endl;

		// Mark protein as done
		done[protein_id] = true;

		// Init todo list with this protein
		stack<unsigned int> todo;				// Coloring stack
		todo.push(protein_id);

		// Collect members of the current connected component
		vector<unsigned int> current_group;			// Keep track of group
		while (!todo.empty()) {
			// Get next protein & remove it from stack
			unsigned int next = todo.top();
			todo.pop();
			// Add protein to current group
			current_group.push_back(next);

			// Add targets to todo list
			for (unsigned int i = 0; i < graph[next].edges.size(); i++) {
				unsigned int target = graph[next].edges[i];
				// Mark & add if unknown yet
				if (!done[target]) {
					done[target] = true;
					todo.push(target);
				}
			} // For each target
		} // While todo list is not empty

		// Do not report singles
		if (current_group.size() < 2) {
			protein_id++;
			continue;
		}

		// Connectivity analysis
		if (debug_level > 0) cerr << getTime() << " [DEBUG]   Found connected component: " << current_group.size() << " proteins (ID: " << protein_id << ")" << endl;

		// Skipp those that are too large
		if (current_group.size() > param_max_nodes) {
			cerr << " [WARNING]  Skipping connected component that contains " << current_group.size() << " elements. This behavior can be adjusted using -maxnodes, still you better consider a higher e-value or coverage thresholds." << endl;
			clear_edges(current_group);
			continue;
		}

		double connectivity = getConnectivity(current_group);
		if (debug_level > 0) cerr << getTime() << " [DEBUG]   Connectivity was " << connectivity << endl;

		// negative values indicate an overwrite. con will be accepted due to good average species coverage
		if (param_con_threshold && connectivity >= 0 && connectivity < param_con_threshold) {
			if (debug_level > 0) cerr << getTime() << " [DEBUG]   Reiterating" << endl;
			// Split groups is done in getConnectivity function
			// Remove flags
			for (unsigned int i = 0; i < current_group.size(); i++) done[current_group[i]] = false;
			continue;
		}
	
		// Output
		if 	(connectivity < 0) {print_group(current_group,-connectivity);}				// Connectivity threshold not used
		else if (connectivity >= param_con_threshold) {print_group(current_group,connectivity);}

		// Print stats
		stats(protein_id,protein_counter);
		
		// Clean up and go on with the next connected component
		clear_edges(current_group);
		protein_id++;
	} // For each protein (mainloop)
	stats(1,1);
	if (param_verbose) cerr << "\r" << "Done                       " << endl;
}


///////////////////////////////////////////////////////////
// Basic Graph functions
///////////////////////////////////////////////////////////
// Remove all edges from the given list of protein ids -- openMP A ML
void clear_edges(vector<unsigned int>& nodes) {
//	#pragma omp parallel for
	for (unsigned int i = 0; i < nodes.size(); i++) graph[nodes[i]].edges.clear();
}

// Remove edges that go beyond the given group groups a, works in a single direction, no overlap in parallel runs if lists are without overlap [TRY]
void removeExternalEdges(map<unsigned int,bool>& a) {
	// For each node id in Hash
	for (map<unsigned int,bool>::iterator it = a.begin(); it != a.end(); it++) {
		unsigned int node = it->first;
		// For each edge of this node
		for (unsigned int j = 0; j < graph[node].edges.size();) {
			// If target node is not our group, remove
			if (a.find(graph[node].edges[j]) == a.end()) {
				remove_edge_index(node,j);				// Works in a single direction, has to be called for outsider group as well
			}
			else {
				j++; // only increment if all is fine
			}
		}
	}
}

// Remove edge in a single direction given an index -- openMP safe
void remove_edge_index(const unsigned int node_id, const unsigned int index) {
	protein node = graph[node_id];
	if (index >= graph[node_id].edges.size()) {
		stringstream ss;
		ss << "Called index out of bounds: " << node.full_name << " sized " << graph[node_id].edges.size() << " called " << index;
		throw ss.str();
	}

	// log
//	cerr << "Removed " << node.full_name << " -> " << graph[node.edges[index].edge].full_name << " (weight: " << node.edges[index].weight << ")" << endl;
//	#pragma omp critical
	graph_clean << node.full_name << "\t" << species[node.species_id] << "\t" << graph[node.edges[index]].full_name << "\t" << species[graph[node.edges[index]].species_id] << endl;

//	print_edgelist(node, index, node_id);

	// overwrite this position with the last
	graph[node_id].edges[index] = graph[node_id].edges.back();
	// and remove last (also works if index == last)
	graph[node_id].edges.pop_back();

//	node = graph[node_id];
//	print_edgelist(node, index, node_id);
}

///////////////////////////////////////////////////////////
// File parser
///////////////////////////////////////////////////////////
void parse_file(string file) {
	if (param_verbose) cerr << "Reading " << file << endl;
	string line;
	ifstream graph_file(file.c_str());
	if (graph_file.is_open()) {
		// For each line
		string file_a = "";	unsigned int file_a_id = 0;
		string file_b = "";	unsigned int file_b_id = 0;
		while (!graph_file.eof()) {
			getline(graph_file, line);
			vector<string> fields;
			tokenize(line, fields, "\t");
			// Header line
			if (fields.size() == 2 && fields[0].substr(0, 1) == "#") {
				file_a = fields[0].substr(2, fields[0].size()-2);
				file_b = fields[1];

				if (file_a == "file_a" && file_b == "file_b") continue;	// Init Header

				// Map species a
				if (species2id.find(file_a) == species2id.end())	{
						species.push_back(file_a);
//						cerr << "Species " << species_counter << ": " << file_a << endl;
						species2id[file_a] = species_counter++;
				}
				// Map species b
				if (species2id.find(file_b) == species2id.end())	{
//						cerr << "Species " << species_counter << ": " << file_b << endl;
						species.push_back(file_b);
						species2id[file_b] = species_counter++;
				}

				file_a_id = species2id[file_a];
				file_b_id = species2id[file_b];
			}
			// Data line
			else if ((fields.size() == 6 || fields.size() == 8) && fields[0].substr(0, 1) != "#") {
				// a b e1 b1 e2 b2 score

//				cerr << protein_counter << ": " << fields[0] << " <-> " << fields[1] << endl;

				// 5.16 deal with duplicated IDs by adding file ID to protein ID
				string ida = fields[0];
				string idb = fields[1];
				fields[0] += " " + file_a_id;
				fields[1] += " " + file_b_id;

				// 5.16 do not point to yourself
				if (!fields[0].compare(fields[1])) {continue;}

				// A = new protein
				if (protein2id.find(fields[0]) == protein2id.end())	{
					protein a;
					a.full_name	= ida;
					a.species_id	= file_a_id;
					protein2id[fields[0]] = protein_counter++;
					graph.push_back(a);
				}
				// B = new protein
				if (protein2id.find(fields[1]) == protein2id.end())	{
					protein b;
					b.full_name	= idb;
					b.species_id	= file_b_id;
					protein2id[fields[1]] = protein_counter++;
					graph.push_back(b);
				}

		
				// Add link to graph (reciprocal)					
				unsigned int a_id = protein2id[fields[0]];
				unsigned int b_id = protein2id[fields[1]];

				// 5.17, add weight
				graph[a_id].edges.push_back(b_id);
				graph[b_id].edges.push_back(a_id);
				edges++;
			}
		}
	}
	else {
		throw string("Could not open file " + file);
	}
}

///////////////////////////////////////////////////////////
// Output
///////////////////////////////////////////////////////////
// Sort
void sort_species(void) {
	reorder_table.reserve(species_counter);
	vector<string> species_sorted (species_counter);
	copy(species.begin(), species.end(), species_sorted.begin());
	sort(species_sorted.begin(), species_sorted.end());
	// find new locations (not efficent but list is small)	
	for (unsigned int i = 0; i < species_counter; i++) {
		for (unsigned int j = 0; j < species_counter; j++) {
			if (species[i] == species_sorted[j]) {reorder_table[j] = i; continue;}
		}
	}
}


// Progress stats
void stats(double i, double size) {
	if (!param_verbose) return;
	double stat = double(i/size*100);
	if (last_stat * 1.01 < stat) {
		last_stat = stat;
		cerr << "\r" << "                          ";
		cerr << "\r" << "Clustering: " << setprecision (2) << fixed << stat << "%";
//		cerr << "Clustering: " << stat << "%" << endl;
	}
}

// Header with species names
void print_header() {
	cout << "# Species\tGenes\tAlg.-Conn.";
	for (unsigned int i = 0; i < species_counter; i++) {
		cout << "\t" << species[reorder_table[i]];
	}
	cout << endl;
}

// Group formatting
void print_group(vector<unsigned int>& nodes, double connectivity) {
	vector<string> line(species_counter,"*");	// Output vector

	unsigned int species_number = 0;
	// For each protein in group
	for (unsigned int i = 0; i < nodes.size(); i++) {
		unsigned int current_protein = nodes[i];
		unsigned int current_species = graph[current_protein].species_id;
		if (line[current_species].compare("*") == 0) {
			line[current_species] = graph[current_protein].full_name;
			species_number++;
		}
		else {
			line[current_species].append(","+graph[current_protein].full_name);
		}
	}

	cout << species_number << "\t" << nodes.size() << "\t" << setprecision (3) << connectivity;

	// List group data
	for (unsigned int i = 0; i < species_counter; i++) {
		
		// sort line if multi protein
		vector<string> fields;
		tokenize(line[reorder_table[i]], fields, ",");
		if (fields.size() > 1) {
			sort(fields.begin(), fields.end());
			line[reorder_table[i]] = fields[0];
			for (unsigned int k = 1; k < fields.size(); k++) {
				line[reorder_table[i]].append(","+fields[k]);
			}
		}

		// output
		cout << "\t" << line[reorder_table[i]];
	}
	cout << endl;
}

// Calculate number of species formatting
float calc_group(vector<unsigned int>& nodes) {
	vector<unsigned int> line(species_counter,0);	// Species vector
	unsigned int species_number = 0;
	// For each protein in group, count species
	for (unsigned int i = 0; i < nodes.size(); i++) {
		unsigned int current_protein = nodes[i];
		unsigned int current_species = graph[current_protein].species_id;
		if (line[current_species] == 0) {species_number++;}			// new species
		line[current_species]++;						// calc
	}
	line.clear();

	unsigned int sum = 0;
	for (unsigned int current_species = 0; current_species < species_counter; current_species++) {
		sum += line[current_species];
	}
	
	float avg = float(sum)/float(species_number);
	return avg;
}


///////////////////////////////////////////////////////////
// Misc functions
///////////////////////////////////////////////////////////
// Convert string to double
double string2double(string str) {
	istringstream buffer(str);
	double value;
	buffer >> value;
	return value;
}


// Split a string at a certain delim
void tokenize(const string& str, vector<string>& tokens, const string& delimiters = "\t") {
	// Skip delimiters at beginning.
	string::size_type lastPos = str.find_first_not_of(delimiters, 0);
	// Find first "non-delimiter".
	string::size_type pos = str.find_first_of(delimiters, lastPos);

	while (string::npos != pos || string::npos != lastPos) {
		// Found a token, add it to the vector.
		tokens.push_back(str.substr(lastPos, pos - lastPos));
		// Skip delimiters.  Note the "not_of"
		lastPos = str.find_first_not_of(delimiters, pos);
		// Find next "non-delimiter"
		pos = str.find_first_of(delimiters, lastPos);
	}
}

///////////////////////////////////////////////////////////
// Algebraic connectivity functions
///////////////////////////////////////////////////////////
// Return maximum degree of given protein_ids -- openMP A ML
unsigned int max_deg(vector<unsigned int>& nodes) {
	unsigned int max = 0;

//	#pragma omp parallel for reduction(max: max)
	for (unsigned int i = 0; i < nodes.size(); i++) {
		unsigned int degree = graph[nodes[i]].edges.size();
		if (degree > max) max = degree;
	}
	return max;
}

// Generate random vector x of size size	-- openMP C ML PLZ Check!
vector<double> generate_random_vectorGOOD(const unsigned int size) {
	vector<double> x(size);

	x[0] = (double)(rand() % 999+1)/1000;	// 0 to 1
	for (unsigned int i = 1; i < size; i++) {
		x[i] = (double)(rand() % 999+1)/1000;	// 0 to 1
		if (x[i] == x[i-1]) x[i] /= 3;		// Check: at least one value must be different from the others but still within 0 and 1
	}

	return x;
}

// Generate random vector x of size size	-- openMP C ML PLZ Check!
vector<double> generate_random_vector(const unsigned int size) {
	vector<double> x(size);

	for (unsigned int i = 0; i < size; i++) {
		x[i] = i+1;
	}

	return x;
}

// determine new X, Formula (1) -- openMP B ML
vector<double> get_new_x(vector<double> x, vector<unsigned int>& nodes, map<unsigned int,unsigned int>& mapping) {
	vector<double> x_new(x.size());
	// go through all nodes (rows of A)

//	#pragma omp parallel for
	for (unsigned int i = 0; i < nodes.size(); i++) {
//		unsigned int node = nodes[i];	// node x is node y here
		x_new[i] = 0;
		// go through adjacency list of node (cols of A_i)
		for (unsigned int j = 0; j < graph[nodes[i]].edges.size(); j++) {
			// y points to z, so take entry z from x
			unsigned int abs_target = graph[nodes[i]].edges[j];
			unsigned int rel_target = mapping[abs_target];
			// -> Sum over x entries
			x_new[i] += x[rel_target];
		}
	}
	return x_new;
}

// Make vector x orthogonal to 1, Formula (2) -- openMP A ML
vector<double> makeOrthogonal(vector<double> x) {
	double sum = 0;

//	#pragma omp parallel for reduction(+: sum)
	for (unsigned int i = 0; i < x.size(); i++) {sum += x[i];}

	double average = sum/x.size();

//	#pragma omp parallel for
	for (unsigned int i = 0; i < x.size(); i++) {x[i] -= average;}

	return x;
}

// Normalize vector x, Formula (4) -- openMP A ML
vector<double> nomalize(vector<double> x, double *length) {
	double sum = 0;

//	#pragma omp parallel for reduction(+: sum)
	for (unsigned int i = 0; i < x.size(); i++) {sum += x[i]*x[i];}

	*length = sqrt(sum);
	if (*length == 0) {*length = 0.000000001;}	// ATTENTION not 0!

//	#pragma omp parallel for
	for (unsigned int i = 0; i < x.size(); i++) {x[i] /= *length;}
	return x;
}

// Qx, Formula (5) -- openMP A ML
vector<double> getY(double max_degree, vector<double> x_hat, vector<double> x_new, vector<unsigned int>& nodes){
	// (2*maxdeg - grad_node_i ) * x_hat_i + new_x_i

//	#pragma omp parallel for
	for (unsigned int i = 0; i < x_hat.size(); i++) {
		x_hat[i] *= (2*max_degree - graph[nodes[i]].edges.size());
		x_hat[i] += x_new[i];
	}
	return x_hat;
}

unsigned int sumOutDegs(const vector<unsigned int>& nodes) {
	unsigned int sum = 0;
	for (unsigned int a = 0; a < nodes.size(); a++) {
		unsigned int from = nodes[a];
		sum += graph[from].edges.size();
	}
	return sum/2;
}


double getConnectivity(vector<unsigned int>& nodes) {
	// Special cases quicky
	if (nodes.size() == 2) {return 1;}

	if (nodes.size() == 3) {
		if (graph[nodes[0]].edges.size() == 2 && graph[nodes[1]].edges.size() == 2) return 1;
		else {
			double conn = 1.0/3.0;
			if (conn < param_con_threshold) {
				if (param_min_species != 0) {
					float avg = calc_group(nodes);
					if (debug_level > 0) cerr << getTime() << " [DEBUG]   Found " << avg << " genes/species on average. User asked for " << param_min_species << endl;
					if (avg <= param_min_species) return -conn;
				}
				clear_edges(nodes);	// destroy group
			}		
			return conn;
		}
	}

	if (nodes.size() == 4) {if (graph[nodes[0]].edges.size() == 3 && graph[nodes[1]].edges.size() == 3 && graph[nodes[2]].edges.size() == 3) return 1;}

	// Get max degree of nodes
	unsigned int max_degree = max_deg(nodes);
	
	// Compress value range / map data
	map<unsigned int,unsigned int> mapping;
	for (unsigned int i = 0; i < nodes.size(); i++) {mapping[nodes[i]] = i;}
	
	// Init randomized x 
	vector<double> x = generate_random_vector(nodes.size());

	// Orthogonalize + normalize vector + get initial lenght
	double last_length, current_length = 0;
	vector<double> x_hat = makeOrthogonal(x);
	vector<double> norm = nomalize(x_hat, &last_length);

	// Repeat until difference < epsilon
	unsigned int iter = 0;							// catch huge clustering issues by keeping track here
	
	while(++iter < max_iter) { 
		if (debug_level > 2 && iter%100 == 0) cerr << getTime() << " [DEBUG L2] Step " << iter << " / " << max_iter << endl;
		last_length = current_length;

		// Get a new x
		x = get_new_x(norm, nodes, mapping);
		// Get y
		vector<double> y = getY(max_degree,norm,x,nodes);
		// Orthogonalize
		x_hat = makeOrthogonal(y);
		// Get lenght (lambda) & normalize vector
		norm = nomalize(x_hat, &current_length);
//		if (nodes.size() < 6) {
//		cerr << "I " << iter << ": " << abs(current_length-last_length) << " {" << current_length << "}" << endl; 
//		}
		if (abs(current_length-last_length) < epsilon && iter >= min_iter) break;	// prevent convergence by chance, converge to epsilon
	}

///		cerr << nodes.size() << " nodes done after " << iter << " iterations" << endl;
	if (debug_level > 0) cerr << getTime() << " [DEBUG]   " << iter << " / " << max_iter << " iterations required (error is " << abs(current_length-last_length) << ")" << endl;

	double connectivity = (-current_length+2*max_degree)/(nodes.size());

		//
//		if (nodes.size() < 6) {
//		cerr << "[";
//		for (unsigned int f = 0; f < x_hat.size(); f++) {
//			cerr << x_hat[f] << ", ";
//		}
//		cerr << "]" << endl;
//		}
		//

	// 5.17 catch hardly-converging groups
	if (iter >= max_iter) {
		if (debug_level > 0) cerr << getTime() << " [DEBUG]   Connectivity score of connected component with " << nodes.size() << " elements did not converge perfectly in time." << endl;
	}

//	debug__conn_integrity(nodes,connectivity);

	// Split groups if connectivity is too low, remove tree like structures that might have arosen
	if (connectivity < param_con_threshold) {
		
		// 5.17 new threshold option overwrites connectivity
		if (param_min_species != 0) {
			float avg = calc_group(nodes);
			if (debug_level > 0) cerr << getTime() << " [DEBUG]   Found " << avg << " genes/species on average. User asked for " << param_min_species << endl;
			if (avg <= param_min_species) {
				if (debug_level > 0) cerr << getTime() << " [DEBUG]   Group is going to be accepted despite connectivity " << connectivity << endl;
				// just to be safe from infinit loops due to rounding
				if (connectivity == 0) connectivity = 0.001;
				// no splitting despite bad connectivity
				return -connectivity;			
			}
		}
///		cerr << "Conn: " << connectivity << endl; ///
		splitGroups(x_hat, nodes);
	}
	
	return connectivity;
}

// Split connected component according to eigenvector -- openMP B ML
void splitGroups(vector<double>& y, vector<unsigned int>& nodes){

//	debug__graph_integrity(nodes);

	// Store data about two groups (Zero cannot be assigned with certainty)
	map<unsigned int,bool> groupA, groupB, groupZero;

	// Beware of MP an Maps!
//	#pragma omp parallel for
	for (unsigned int i = 0; i < y.size(); i++) {
		if      (abs(y[i]) < param_sep_purity)		{
//			#pragma omp critical
			groupZero[nodes[i]] = true;
//			if (debug_level > 2) cerr << i << ":" << y[i] << " remove" << endl;
		}
		else if (y[i] < 0) 				{
//			#pragma omp critical
			groupA[nodes[i]] = true;
//			if (debug_level > 2) cerr << i << ":" << y[i] << " groupA" << endl;
		}
		else              				{
//			#pragma omp critical
			groupB[nodes[i]] = true;
//			if (debug_level > 2) cerr << i << ":" << y[i] << " groupB" << endl;
		}
	}

	// Catch error in laplacien calcs
	if ((groupA.size() == 0 || groupB.size() == 0) && groupZero.size() == 0) {
		throw string("[CRITICAL ERROR]   Failed to partition subgraph! This might lead to an infinit loop. Please submit the .blastgraph file to lechner@staff.uni-marburg.de to help fixing this issue.");
	}


//	#pragma omp parallel sections
	{
//		#pragma omp section
		{removeExternalEdges(groupZero);}
//		#pragma omp section		
		{removeExternalEdges(groupA);}
//		#pragma omp section
		{removeExternalEdges(groupB);}
	}


//	debug__graph_integrity(nodes);

}

string getTime(void) {
	time_t t = time(0);   // get time now
	struct tm * now = localtime( & t );
	std::ostringstream oss;
	if (now->tm_hour < 10) oss << "0";
	oss << now->tm_hour << ":";
	if (now->tm_min < 10) oss << "0";
	oss << now->tm_min << ":";
	if (now->tm_sec < 10) oss << "0";
	oss << now->tm_sec;
	return oss.str();
}


////////////////////// Debug
void debug__print_edgelist (protein& node, const unsigned int index, const int node_id) {
	cerr << node_id << ": ";	
	for (unsigned int j = 0; j < node.edges.size(); j++) {
		if (j == index) cerr << "*";
		cerr << node.edges[j] << " ";
	}
	cerr << endl;
}

void debug__conn_integrity(vector<unsigned int>& nodes, double conn) {
	if (nodes.size() > 5) return;

	unsigned int sum = 0;
	for (unsigned int a = 0; a < nodes.size(); a++) {
		unsigned int from = nodes[a];
		sum += graph[from].edges.size();
	}

	sum /= 2;

	if (nodes.size() == 3) {
		if (sum == 2 && conn > 0.4) {
			cerr << "gs 3 with 2 edges had conn " << conn << endl;
			throw string("integrity issue: connectivity of group size 3a");
		}
		if (sum == 3 && conn < 1) {
			cerr << "gs 3 with 3 edges had conn " << conn << endl;
			throw string("integrity issue: connectivity of group size 3b");
		}
	}
	if (nodes.size() == 4) {
		if (sum == 3 && conn > 0.4) {
			cerr << "gs 4 with 3 edges had conn " << conn << endl;
			throw string("integrity issue: connectivity of group size 4a");
		}
		if (sum < 6 && conn == 1) {
			cerr << "gs 4 with <6 edges had conn " << conn << endl;
			throw string("integrity issue: connectivity of group size 4b");
		}
		if (sum == 6 && conn < 1) {
			cerr << "gs 4 with 6 edges had conn " << conn << endl;
			throw string("integrity issue: connectivity of group size 4c");
		}
	}
	if (nodes.size() == 5) {
		if (sum == 4 && conn > 0.4) {
			cerr << "gs 5 with 4 edges had conn " << conn << endl;
			throw string("integrity issue: connectivity of group size 5a");
		}
		if (sum < 10 && conn == 1) {
			cerr << "gs 5 with <10 edges had conn " << conn << endl;
			throw string("integrity issue: connectivity of group size 5b");
		}
		if (sum == 10 && conn < 1) {
			cerr << "gs 5 with 10 edges had conn " << conn << endl;
			throw string("integrity issue: connectivity of group size 5c");
		}
	}
}

void debug__graph_integrity(vector<unsigned int>& nodes) {
	// For each node
	for (unsigned int a = 0; a < nodes.size(); a++) {
		unsigned int from = nodes[a];
//		cerr << "From: " << from << endl;

		// For each edge in + direction
		for (unsigned int i = 0; i < graph[from].edges.size(); i++) {
			unsigned int to = graph[from].edges[i];
//			cerr << " To:  " << to << endl;

			if (to == from) {cerr << "ERROR: Edge from " << from << " to " << to << " is selfevident" << endl; throw string("integrity issue self hit");}

			// Check reverse direction
			// Foreach edge in - direction
			bool found = false;
//			cerr << " Back:";
			for (unsigned int j = 0; j < graph[to].edges.size(); j++) {
				unsigned int back = graph[to].edges[j];
//				cerr << " " << back;
				if (back == from) {found = true; continue;}
			}
//			cerr << endl;
			if (!found) {
				cerr << "ERROR: Edge from " << from << " to " << to << " is unidirectional" << endl; throw string("integrity issue direction");
			}
		}
	}
}


